"""Flow Folder

Collection of classes, functions and methods to describe flow inside pipes and jetpumps.
Flow equations for single phase, two phase, reservoir inflow and jet pumps are present.
"""

from .inflow import InFlow
